//eval.h

#ifndef _EVAL_H_
#define _EVAL_H_

#include <iostream>
#include <cstdlib>
#include <cstring>
#include <ctime>

#include <stack>
#include <vector>
#include <utility>

using namespace std;

#include "unit.h"
#include "flib.h"
#include "common.h"
#include "slib.h"

//!Struktura przechowuje opis operatora binarnego
typedef struct _TOperator
{
	char *opstr;		//!Ciag znakow operatora
	char *fname;		//!Nazwa funkcji dwuargumentowej odpowiadajaca operatorowi
	int pr;				//!Priorytet operatora
} TOperator;



//!Klasa umozliwiajaca obliczanie wartosci wyrazen
template<class Type>
class CEval
{
private:
//pola prywatne
	CBaseUnit<Type> *root;		//korzen drzewa wyrazen

	
	static CFLib stdlib;		//standardowa biblioteka zmiennych i funkcji
								//niezalezna od objektu
	CFLib		 extlib;		//rozszerzona biblioteka zmiennych i funkcji
								//konkretna dla objektu

	char *expression;			//zapamietane wyrazenie jako ciag znakow

	vector<char*> local_names;					//nazwy zmiennych lokalnych
	vector<Type*> local_ptr;					//adresy do tych zmiennych

	vector<Type*> global_ptr;					//adresy do zmiennych globalnych (tworzone przez AddVar)
												//pamietamy je, aby usunac je na koncu istnienia obiektu

//metody prywatne
	void reset(const char *s="");				//zwalnia pamiec na korzen i ciag identyfikujacy wyrazenie

	void dellocals();							//usuwa zmienne lokalne z extlib

	TNodeElement *findoperator(char *,int&);	//wyszukuje dany operator

	int getchartype(int,int);					//zwraca typ znaku
												//wedlug ponizszego 'enuma'

	enum CharTypes
	{
		CT_Null,
		CT_Number,
		CT_Operator,
		CT_LBracket,	//(
		CT_RBracket,	//)
		CT_Comma,		//,
		CT_Name
	};

	char *getelement(int&,char*,int&,int);		//zwraca pojedynczy element wyrazenia w postaci ciagu znakow
												//albo zero, jesli doszlismy do konca wyrazenia
												//dodatkowo w ostatnim parametrze zwraca typ elementu
												//wymagany jest brak odstepow w wyrazeniu

	int unitfromnodeelement(TNodeElement *,		//tworzy nowy objekt CBaseUnit w zaleznosci
		stack< CBaseUnit<Type>* >&);			//od rodzaju TNodeElement i daje go na stos (2 param.)

	//priorytety wewnetrzne
	enum Priorities
	{
		Priority_Bracket=1000,
		Priority_Comma=900
	};

	void init();								//initializuje klase
	void copy(CEval &);							//kopiuje elementy


public:
	CEval();
	~CEval();

	//!Konstruktor kopiujacy
	/*!
	 * Tworzy nowe drzewo wyrazen na podstawie objektu podanego w parametrze.
	 */
	CEval(CEval &);

	//!Operator "="
	/*!
	 * Tworzy nowe drzewo wyrazen  na podstawie objektu podanego w parametrze.
	 */
	CEval &operator=(CEval &);

	//!Generuje drzewo wyrazenia na podstawie ciagu znakow
	/*!
	 * W parametrze przekazujemy wyrazenie jako ciag znakow.
	 */
	int Set(const char *);

	//!Generuje drzewo wyrazenia na podstawie ciagu znakow
	/*!
	 * W pierwszym parametrze przekazujemy wyrazenie.
	 * Ponadto tworzy zmienne lokalne o nazwach podanych w wektorze.
	 */
	int Set(const char *,vector<char*>&);

	//!Oblicza wartosc wyrazenia
	inline Type operator()();

	//!Oblicza wartosc wyrazenia
	/*!
	 * Ponadto przekazujemy w wektorze wartosci zmiennych zadeklarowanych w funkcji Set.
	 */
	Type operator()(vector<Type>&);

	//!Tworzy i dodaje zmienna globalna (do extlib)
	/*!
	 * W parametrach przekazujemy kolejno nazwe i wartosc poczatkowa zmiennej.
	 */
	inline int AddVar(const char*,Type);

	//!Zwraca adres do zmiennej globalnej (z extlib)
	inline Type *GetVarPtr(const char*);

	//!Usuwa zmienna globalna (do extlib)
	inline void DelVar(const char*);

	//!Wypisuje dostepne nazwy zmiennych/funkcji
	const char *GetNameList() 
	{
		static string s;
		s=stdlib.NameList();
		s.append(extlib.NameList());
		return s.c_str();
	}

};


//implementacje


//Zmienne statyczne

//lokalna tablica przechowuje identyfikatory obslugiwanych operatorow i nazw funkcji im odpowiadajace
//ostanie pole okresla priorytet operatora
//funkcje musza byc zadeklarowane w bibliotece standardowej programu
static TOperator OperatorTab[]=
{
	{"+","ADD",200},
	{"-","SUB",200},
	{"*","MUL",100},
	{"/","DIV",100},
	{"^","POW",80},
};

//!ilosc elementow tablicy OperatorTab
#define OPERATORTAB_SIZE (sizeof(OperatorTab)/sizeof(TOperator))


//Pola statyczne

template<class Type>
CFLib CEval<Type>::stdlib(0);	//biblioteka funkcji standardowych, jedna instancja na dany typ
								//parametr dla konstruktora ma na celu oszukanie 'dynamic initializer'


//Metody prywatne

//str	-	ciag okreslajacy wyrazenie
template<class Type>
void CEval<Type>::reset(const char *str)
{
	static CValUnit<Type> _0(0);
	if (root)
	{
		if (root!=&_0)
		delete root;
		root=&_0;		//przypisujemy korzeniowi unit zwracajacy 0
	}
	else
	root=&_0;

	const char *nullstr="";
	if (expression)
	//zainicjowalismy juz ciag expression
	{
		if (expression!=nullstr) //usuwamy tylko w przypadku wczesniejszej alkocji
		delete []expression;

		int len=strlen(str)+1;
		if (len>1)
		{
			expression=new char [len];		//chcemy miec pamiec na nowy ciag znakow
			memcpy(expression,str,len);
		}
		else
		expression=(char*)nullstr;	//na ciag ""
	}
	else
	expression=(char*)nullstr;
}

template<class Type>
void CEval<Type>::dellocals()
{
	if (!local_names.empty())
	//jesli kontener napisow nie jest pusty, to usun wczesniej stworzone zmienne lokalne (z extlib)
	{
		vector<char*>::iterator it;
		for (it=local_names.begin(); it!=local_names.end(); ++it)
		{
			//usun element z extlib
			extlib.Del(*it);
		}
	}

	if (!local_ptr.empty())
	//jesli kontener adresow nie jest pusty, to usun wczesniej stworzone zmienne lokalne (z pamieci)
	{
		vector<Type*>::iterator it;
		for (it=local_ptr.begin(); it!=local_ptr.end(); ++it)
		{
			//usun element
			delete *it;
		}
	}

	local_names.clear();
	local_ptr.clear();	//czyscimy wektor adresow
}

//op	-	identyfikator operatora
//pr	-	zwracany priorytet operatora
template<class Type>
TNodeElement *CEval<Type>::findoperator(char *op,int &pr)
{
	try {

	for (int i=0; i<OPERATORTAB_SIZE; i++)
	{
		if (strcmp(op,OperatorTab[i].opstr)==0)
		//znaleziono!
		{
			TNodeElement *n;
			n=stdlib.Find(OperatorTab[i].fname);
			if (n)
			//funkcja istnieje!
			{
				pr=OperatorTab[i].pr;
				return n;
			}
			else
			{
				throw "Operator has not reffunc";
				return 0;
			}
		}
	}

	throw "Undefined operator";
	} //try
	catch(...) {throw;} //przerzuc wyjatek dalej

	return 0;
}


//ch	-	okreslony znak
//pch	-	poprzedni znak (moze byc 0)
template<class Type>
int CEval<Type>::getchartype(int ch,int pch)
{
	if (ch==0) return CT_Null;

	if ((ch=='-') && ((pch==0)||(pch=='(')) )
	//liczba zaczynajaca sie od znaku '-'
	{
		return CT_Number;
	}

	if ((((ch>='0')&&(ch<='9'))||(ch=='.')) && (getchartype(pch,0)!=CT_Name))
	//liczba
	{
		return CT_Number;
	}

	if (((ch>='0')&&(ch<='9')) && (getchartype(pch,0)==CT_Name))
	//nazwa (czesc koncowa moze zawierac cyfry)
	{
		return CT_Name;
	}

	if ( ((ch>='a')&&(ch<='z')) || ((ch>='A')&&(ch<='Z')) )
	//nazwa
	{
		return CT_Name;
	}

	if (ch=='(') return CT_LBracket;
	if (ch==')') return CT_RBracket;
	if (ch==',') return CT_Comma;
	if (ch=='.') return CT_Null;	//blad skladni

	//za operator traktujemy kazdy inny znak 'powyzej' spacji
	if (ch>' ')
	return CT_Operator;

	return CT_Null;
}

//i		-	index w str
//str	-	wyrazenie
//type	-	zwraca element wyrazenia zaczynajac od pozycji i; moze to byc operator, liczba lub nazwa
//len	-	dlugosc napisu str
//zwraca 0 gdy doszlismy do konca wyrazenia lub adres na ciag znakow indentyfikujacy element
//zmienia wartosc i
template<class Type>
char *CEval<Type>::getelement(int &i,char *str,int &type,int len)
{
	//bufor pomocniczy w ktorym przechowujemy wyjsciowy ciag znakow
	const int buflen=64;
	static char buf[buflen];
	static int ptype;
	if (i==0) ptype=CT_Null;

	if (i>=len) return 0;

	//pierwszy znak okresla juz typ
	type=getchartype(str[i],i>0? str[i-1]:0);

	int l=1;
	for (++i; i<len; i++)
	{
		int t=getchartype(str[i],str[i-1]);
		if ((t!=type)||(t==CT_LBracket)||(t==CT_RBracket)) break;
		l++;
	}

	if ((ptype==CT_Number)&&((type==CT_Name)||(type==CT_LBracket)))
	//w sytuacji typu 2sin(... czy 2(...
	{
		buf[0]='*';
		buf[1]=0;
		type=CT_Operator;
		ptype=type;
		i-=l;			//wracamy do punktu wyjscia
		return buf;
	}

	ptype=type;

	//kopiuj do bufora dane
	if (l>=buflen) l=buflen;
	
	memcpy(&buf,&str[i-l],l);
	buf[l]=0;
	if ((l==1)&&(buf[0]=='-')&&(type==CT_Number)) {buf[1]='1'; buf[2]=0;} //ciag "-" zamieniamy na "-1"

	return buf;
}

//n	-	element wezla drzewa
//s	-	stos na ktory odkladamy nowopowstaly unit
//zwraca 0 gdy nie poprawne dane
template<class Type>
int CEval<Type>::unitfromnodeelement(TNodeElement *n,stack< CBaseUnit<Type>* > &s)
{
	CBaseUnit<Type> *u1,*u2;
	if (!n) return 1; //nieokreslony element, ale moze to byc nawias
	try {

	switch (n->flags)
	{

	case TNodeElement::Var:		//zmienna
		s.push(new CVarUnit<Type>((Type*)(n->ref)));
		break;

	case TNodeElement::Func1:	//funkcja 1-arg
		if (s.empty()) throw "Syntax error";	//musi byc argument funkcji
		u1=s.top();	//wez argument
		s.pop();
		s.push(new C1OpUnit<Type>(u1,(C1OpUnit<Type>::TFunc)(n->ref)));
		break;

	case TNodeElement::Func2:	//funkcja 2-arg
		if (s.empty()) throw "Syntax error";	//musi byc argument funkcji
		u2=s.top();	//wez argument 2
		s.pop();
		if (s.empty()) throw "Syntax error";	//musi byc argument funkcji
		u1=s.top();	//wez argument 1
		s.pop();
		s.push(new C2OpUnit<Type>(u1,u2,(C2OpUnit<Type>::TFunc)(n->ref)));
		break;

	}

	}//try
	catch(...) {throw;} //przerzucamy wyjatek dalej
	return 1;
}

//inicjacja klasy
template<class Type>
void CEval<Type>::init()
{
	//UWAGA: konstruktor stdlib(0) jest wywolywany potem
	//STANDARDOWE FUNKCJE
	//funkcje operatorow
	stdlib.Add("ADD",(int)(C2OpUnit<Type>::TFunc)&SLib::Add<Type>,TNodeElement::Func2);
	stdlib.Add("SUB",(int)(C2OpUnit<Type>::TFunc)&SLib::Sub<Type>,TNodeElement::Func2);
	stdlib.Add("MUL",(int)(C2OpUnit<Type>::TFunc)&SLib::Mul<Type>,TNodeElement::Func2);
	stdlib.Add("DIV",(int)(C2OpUnit<Type>::TFunc)&SLib::Div<Type>,TNodeElement::Func2);
	stdlib.Add("POW",(int)(C2OpUnit<Type>::TFunc)&SLib::Pow<Type>,TNodeElement::Func2);

	//funkcje matematyczne
	stdlib.Add("ABS",(int)(C1OpUnit<Type>::TFunc)&SLib::Abs<Type>,TNodeElement::Func1);
	stdlib.Add("SGN",(int)(C1OpUnit<Type>::TFunc)&SLib::Sgn<Type>,TNodeElement::Func1);

	stdlib.Add("SIN",(int)(C1OpUnit<Type>::TFunc)&SLib::Sin<Type>,TNodeElement::Func1);
	stdlib.Add("COS",(int)(C1OpUnit<Type>::TFunc)&SLib::Cos<Type>,TNodeElement::Func1);
	stdlib.Add("TAN",(int)(C1OpUnit<Type>::TFunc)&SLib::Tan<Type>,TNodeElement::Func1);
	stdlib.Add("CTAN",(int)(C1OpUnit<Type>::TFunc)&SLib::Ctan<Type>,TNodeElement::Func1);

	stdlib.Add("FLOOR",(int)(C1OpUnit<Type>::TFunc)&SLib::Floor<Type>,TNodeElement::Func1);
	stdlib.Add("CEIL",(int)(C1OpUnit<Type>::TFunc)&SLib::Ceil<Type>,TNodeElement::Func1);

	stdlib.Add("LN",(int)(typename C1OpUnit<Type>::TFunc)&SLib::Ln<Type>,TNodeElement::Func1);

	stdlib.Add("RND",(int)(C1OpUnit<Type>::TFunc)&SLib::Rnd<Type>,TNodeElement::Func1);

	stdlib.Add("SQRT",(int)(C1OpUnit<Type>::TFunc)&SLib::Sqrt<Type>,TNodeElement::Func1);
	srand((unsigned int)time(0));

	//stale matematyczne
	//zawsze sie pojawia ostrzezenia przy pewnych typach, ale konwersja jest swiadoma
#pragma warning(push,1)
	static Type PI=3.1415926535897932384626433832795;
	static Type E =2.7182818284590452353602874713527;
#pragma warning(pop)
	stdlib.Add("PI", (int)&PI,TNodeElement::Var);
	stdlib.Add("E",  (int)&E,TNodeElement::Var);

	//bez ustalonego wyrazenia
	root=0;			//pierwsze wywolania
	expression=0;	//jw.

	reset("");	//ustawiamy odpowiednio pola root i expression
}

//e	-	kopiujemy z tej klasy
template<class Type>
void CEval<Type>::copy(CEval &e)
{
	if (e.local_names.empty())
	Set(e.expression);
	else
	Set(e.expression,e.local_names);
}

//Metody publiczne

//konstruktor domyslny
template<class Type>
CEval<Type>::CEval()
{
	init();
}

//e	-	klasa ktora kopiujemy
template<class Type>
CEval<Type>::CEval(CEval &e)
{
	init();
	copy(e);
}

//e	-	klasa ktora kopiujemy
template<class Type>
CEval<Type> &CEval<Type>::operator=(CEval &e)
{
	if (this!=&e)
	{
		copy(e);
	}
	return *this;
}

template<class Type>
CEval<Type>::~CEval() 
{
	dellocals();			//usuwamy zmienne lokalne
	reset("");				//usuwamy drzewo wyrazen i pole expression

	//usuwamy wszystkie zmienne wytworzone przez AddVar
	vector<Type*>::iterator it;
	for (it=global_ptr.begin(); it!=global_ptr.end(); ++it)
	{
		//usun zmienna
		delete *it;
	}
}

//str	-	parsowany ciag znakow
template<class Type>
int CEval<Type>::Set(const char *str)
{
	stack< CBaseUnit<Type>* > unitstack;
	stack< pair <TNodeElement*, int > > funcstack;
	CBaseUnit<Type> *u;
	TNodeElement *n;

	//zmienne pomocnicze
	char *s;	//przechowujemy wskaznik do ciagu znakow pojedynczego elementu wyrazenia
	int t;		//zwracany typ elementu wyrazenia
	int pr;		//zwracany priorytet
	int i,j;	//zmienne indeksowe

	int isfunc=0;	//okresla, czy poprzedni element byl funkcja

	//usun spacje i niepoprawne znaki z ciagu str
	char *nospace_str;	//ciag znakow powstaly z str przez opuszczenie spacji
	nospace_str=new char[strlen(str)+1];

	i=0; j=0;
	while (str[i]!=0) 
	{
		if ((str[i]>' ')&&(str[i]<128)) nospace_str[j++]=str[i]; 
		i++;
	}
	nospace_str[j]=0;
	
	i=0;	//licznik w nospace_str

	reset(str);		//usun poprzednie drzewo wyrazen

	int len=strlen(nospace_str);

	//PARSOWANIE WYRAZENIA
	try{

	while (s=getelement(i,nospace_str,t,len))
	//okresl typ ciagu s i przetwarzaj go
	{
		switch (t)
		{
		case CT_Number:		//liczba
#pragma warning(push,1)
			u=new CValUnit<Type>(static_cast<Type>(atof(s)));
#pragma warning(pop)
			unitstack.push(u);
			break;

		case CT_Operator:	//operator
			n=findoperator(s,pr);
			//jesli nie znaleziono, to wyjdz
			if (!n) throw "Undefined operator";

			if (n->flags==TNodeElement::Null) throw "Invalid name address";

			//tworz drzewo dopoki na stosie sa funkcje o nizszym rownym priorytecie
			while ((!funcstack.empty()) && (funcstack.top().second<=pr))
			{
				//okresl typ i dodaj na stos unitow
				if(!unitfromnodeelement(funcstack.top().first,unitstack)) throw "Syntax error";
				funcstack.pop();
			}

			//dodaj operator na stos funkcji z danym priorytetem
			funcstack.push(pair<TNodeElement*, int>(n,pr));
			break;

		case CT_Name:		//zmienna/funkcja
			//szukaj funkcji/zmiennej w standardowej bibliotece
			n=stdlib.Find(s);
			if (!n)
			//jesli nie znaleziono, to szukaj w bibliotece dodatkowej
			{
				n=extlib.Find(s);
				if (!n)
				//brak takiej nazwy
				{
					throw "Undefined name";
				}
			}

			if (n->flags==TNodeElement::Null) throw "Invalid name address";

			//czy jest to stala badz zmienna
			if (n->flags==TNodeElement::Var)
			{
				//okresl typ i dodaj na stos unitow
				if(!unitfromnodeelement(n,unitstack)) throw "Syntax error";
			}
			else
			//funkcja
			{
				//dodaj funkcje na stos funkcji z priorytetem nawiasu
				funcstack.push(pair<TNodeElement*, int>(n,Priority_Bracket));
				isfunc=1;
			}

			break;

		case CT_LBracket:	//lewy nawias '('
			//jesli poprzedni element byl funckja, to nie bierz pod uwage nawiasu
			if (isfunc) isfunc=0;
			else
			//dodaj nawias na stos funckji, jako znacznik nawiasu
			{
				funcstack.push(pair<TNodeElement*, int>(0,Priority_Bracket));
			}

			break;

		case CT_RBracket:	//prawy nawias ')'
			//'zdejmuj' ze stosu funkcji az do nawiasu badz funkcji
			while ((!funcstack.empty()) && (funcstack.top().second<=Priority_Bracket))
			{
				//okresl typ i dodaj na stos unitow
				if (!unitfromnodeelement(funcstack.top().first,unitstack)) throw "Syntax error";
				if (funcstack.top().second==Priority_Bracket) //znacznik nawiasu
				{
					funcstack.pop();
					break; 
				}
				funcstack.pop();
			}
			break;

		case CT_Comma:		//przecinek ','
			//dodaj na stos funckji znacznik przecinka
			funcstack.push(pair<TNodeElement*, int>(0,Priority_Comma));
			break;
		}
	}

	if (unitstack.empty()) throw "Syntax error";	//pusty stos unitow oznacza blad skladni

	//na koniec zdejmij ze stosu funkcji wszystkie zmienne i dokoncz tworzenie drzewa
	while (!funcstack.empty())
	{
		//okresl typ i dodaj na stos unitow
		if (!unitfromnodeelement(funcstack.top().first,unitstack)) throw "Syntax error";
		funcstack.pop();
	}

	} //try
	catch(const char *s)
	{
		ERR(s);
		delete []nospace_str;
		return 0;
	}
	catch(...) {delete []nospace_str; ERR("Undefined error"); return 0;}

	delete []nospace_str;

	root=unitstack.top();	//korzen powinien byc ostatnim elementem na stosie unitow
	unitstack.pop();
	if (!unitstack.empty()) 
	//jakies elementy na stosie oznaczaja blad skladni
	{
		ERR("Syntax error"); 
		reset(""); 
		return 0;
	}

	//o.k.
	return 1;
}

//str	-	parsowany ciag znakow
//v		-	wektor nazw zmiennych
template<class Type>
int CEval<Type>::Set(const char *str,vector<char*> &v)
{
	//usuwamy poprzednie zmienne lokalne (sa one zawsze tymczasowe, do kolejnego wywolania tej metody)
	dellocals();

	local_names=v;		//kopiujemy wektor napisow

	//dodaj nowe elementy
	vector<char*>::iterator it;
	for (it=local_names.begin(); it!=local_names.end(); ++it)
	{
		Type *l=new Type(0);		//tworz zmienna lokalna = 0
		
		//dodaj ja do biblioteki extlib
		extlib.Add(*it,(int)l,TNodeElement::Var);

		//zapamietaj adres
		local_ptr.push_back(l);
	}

	//parsuj wyrazenie
	return Set(str);
}


template<class Type>
Type CEval<Type>::operator()()
{
	return (*root)();
}

//v	-	wektor wartosci stalych, okreslajacych kolejne wartosci zmiennych lokalnych
template<class Type>
Type CEval<Type>::operator()(vector<Type> &v)
{
	vector<Type>::iterator it;	//dla v
	vector<Type*>::iterator it2=local_ptr.begin();

	//ustal wartosci zmiennych lokalnych na podstawie wektora v
	for (it=v.begin(); it!=v.end(),it2!=local_ptr.end(); ++it,++it2)
	{
		**it2=*it;
	}

	//wywolaj funkcje
	return (*this)();
}

//n	-	nazwa zmiennej
//v	-	wartosc domyslna
template<class Type>
int CEval<Type>::AddVar(const char *n,Type v)
{
	Type *ptr=new Type(v);
	if (!ptr) return 0;
	extlib.Add(n,(int)ptr,TNodeElement::Var);
	global_ptr.push_back(ptr);
	return 1;
}

//n	-	nazwa zmiennej
template<class Type>
Type *CEval<Type>::GetVarPtr(const char *n)
{
	TNodeElement *nd=extlib.Find(n);
	if (nd)
	return (Type*)nd->ref;
	else
	return 0;
}

//n	-	nazwa zmiennej
template<class Type>
void CEval<Type>::DelVar(const char *n)
{
	extlib.Del(n);	//nie usuwamy zmiennej, tylko jej identyfikator w extlib
					//moze byc bowiem tak, ze zmienna bedzie jeszcze wykorzystywana
					//przy obliczaniu wartosci wyrazen
}

#endif //_EVAL_H_