#include <cstring>
#include <list>

#include "flib.h"

using namespace std;

#include "common.h"

//metody prywatne


//c	-	zamieniany znak
int CFLib::char2index(int c) const
{
	if ((c>='A')&&(c<='Z')) return c-'A'+10;
	if ((c>='a')&&(c<='z')) return c-'a'+10;
	if ((c>='0')&&(c<='9')) return c-'0';
	if (c=='_') return 'Z'-'A'+11;
	return -1;
}

//i	-	zamieniana liczba
int CFLib::index2char(int i) const
{
	if ((i>=10)&&(i<=10+'Z'-'A')) return 'A'+i-10;
	if ((i>=0)&&(i<=9)) return i+'0';
	if (i=='Z'-'A'+11) return '_';
	return 0;
}


//s	-	ciag wejsciowy
//t	-	ciag wyjsciowy
//n	-	maksymalna dlugosc ciagu
int CFLib::str2indextab(const char *s,int *t,int n) const
{
	int l=strlen(s);
	if (l>=n) l=n-1;
	try{
	for (int i=0; i<l; i++)
	{
		t[i]=char2index(s[i]);
		if (t[i]==-1) throw "Incorrect name";
	}
	} //try
	catch(char *desc)
	{
		ERR(desc);
		return 0;
	}
	return l;
}

//n	-	wezel
void CFLib::delnode(TNode *n)
{
	if (!n) return;

	for (int i=0; i<FLIB_NCHARS; i++)
	{
		//j=i;
		delnode(n->n[i]);
	}
	if (n->e) delete n->e;
	delete n;
}

//n	-	wezel slownika
//t	-	lista ciagow znakow przechowujaca wzorzec dla zwracanej list funkcji
//np. dla sekwencji AB DEF zwroci funkcje zaczynajace sie od A lub B, a drugi znak musi byc D, E lub F
//pozostale znaki sa dowolne - jeszcze nie zaimplementowane!!!
//it-	okresla czy jest to pierwsze wywolanie funkcji
list<char *> &CFLib::funclist(TNode *n,list<char *> &t,int it)
{
	//pola statyczne
	const int maxbuflen=32;					//rozmiar bufora znakow
	static char currentstring[maxbuflen];	//ciag znakow dodawany do listy


	if (it==0)
	//pierwsze wywolanie
	{
		eraseretlist();						//czysc retlist i dowiazane ciagi znakow
		currentstring[0]=0;
	}

	if (!n) return retlist;
	if (it>=maxbuflen) return retlist; //za dluga nazwa

	if (n->e)	//jesli mamy element w slowniku, to dodajemy go do listy
	{
		currentstring[it]=0;
		char *ch=new char[strlen(currentstring)+1];		//alokuj pamiec na ciag dla retlist
#pragma warning(suppress : 4996)
		strcpy(ch,currentstring);
		retlist.push_back(ch);
	}

	if (t.empty())
	//bierzemy wszystkie znaki
	{
		for (int i=0; i<FLIB_NCHARS; i++)
		{
			if (n->n[i])
			{
				currentstring[it]=index2char(i);
				funclist(n->n[i],t,it+1);
			}
		}
	}
	/*
	else
	//selekcja
	{
		list<char *>::iterator tit=t.begin();
		char *ch=*tit;
		tit++;
		int l=strlen(ch);
		for (int k=0; k<l; k++)
		//dla kazdego znaku
		{
			int i=char2index(ch[k]);
			if (n->n[i])
			{
				currentstring[it+1]=ch[k];
				//funclist(n->n[i],tit,it+1);
			}
		}
	}*/
	currentstring[it]=0;

	return retlist;
}

void CFLib::eraseretlist()
{
	if (!retlist.empty())
	{
		list <char *>::iterator it; //lokalny iterator
		it=retlist.end();
		--it;
		for (; it!=retlist.begin(); --it)
		{
			delete []*it;		//usuwamy nasz ciag znakow
		}
		//pierwszy
		delete []*it;		//usuwamy ostatni ciag znakow
		retlist.clear();	//czyscimy liste
	}
}

//metody publiczne

CFLib::CFLib() 
{
	root=0; 
	nameliststr=0;
}


//name	-	dodawana nazwa
//ref	-	wartosc ref (adres do funkcji/zmiennej/wartosc stala
//flags	-	znaczenie pozycji ref/ilosc argumentow funkcji
TNodeElement *CFLib::Add(const char *name,int ref,int flags)
{
	int itab[FLIB_MAXNAMELEN];						//ciag indeksow

	int l=str2indextab(name,itab,FLIB_MAXNAMELEN);	//w l mamy dlugosc ciagu name
	if (l==0) return 0;								//wyjdz bez dodawania nazwy jesli jest niepoprawna

	TNode *node;	//reprezentuje wezel
	if (!root) 
	{
		root=new TNode;	//alokuj pamiec na korzen
		memset(root,0,sizeof(TNode));
	}
	node=root;

	int j;
	for (int i=0; i<l; i++)
	{
		j=itab[i];
		if (node->n[j]==0)
		//jesli brak kolejnego poddrzewa
		{
			node->n[j]=new TNode;
			memset(node->n[j],0,sizeof(TNode));
		}
		node=node->n[j];
	}

	if (node->e==0)
	node->e=new TNodeElement; //stworz nowy element drzewa
 
	node->e->ref=ref;
	node->e->flags=flags;

	return node->e;
}


//name	-	usuwana nazwa
void CFLib::Del(const char *name)
{
	int itab[FLIB_MAXNAMELEN];						//ciag indeksow

	int l=str2indextab(name,itab,FLIB_MAXNAMELEN);	//w l mamy dlugosc ciagu name
	if (l==0) return;								//wyjdz bez usuwania nazwy jesli jest niepoprawna

	if (!root) return;								//niezainicjowany korzen==brak elementow drzewie

	TNode *node;
	node=root;

	int j;
	for (int i=0; i<l; i++)
	{
		j=itab[i];
		if (node->n[j]==0) return; //oznacza brak takiej nazwy
		node=node->n[j];
	}

	if (node->e)
	{
		//usuwa tylko dany element, konstrukcja drzewa moze sie jeszcze przydac
		//dlatego jej nie niszczymy (nie pozera to duzo pamieci)
		delete node->e;
		node->e=0;
	}
}


//name	-	nazwa szukanego elementu
TNodeElement *CFLib::Find(const char *name)
{
	int itab[FLIB_MAXNAMELEN];						//ciag indeksow

	int l=str2indextab(name,itab,FLIB_MAXNAMELEN);	//w l mamy dlugosc ciagu name
	if (l==0) return 0;								//wyjdz bez usuwania nazwy jesli jest niepoprawna
	if (!root) return 0;							//niezainicjowany korzen==brak elementow drzewie

	TNode *node;
	node=root;

	int j;
	for (int i=0; i<l; i++)
	{
		j=itab[i];
		if (node->n[j]==0) return 0; //oznacza brak takiej nazwy
		node=node->n[j];
	}

	return node->e;
}

//destruktor
CFLib::~CFLib()
{
	delnode(root);
	eraseretlist();
	if (nameliststr) delete [] nameliststr;
	root=0;
	nameliststr=0;
}

char *CFLib::NameList()
{
	if (!root) return "";		//brak elementow

	if (nameliststr) delete [] nameliststr;	//usunmy poprzednia tablice
	nameliststr=0;

	list <char*> l=funclist(root,list<char*>::list());	//pobieramy liste funkcji w postaci listy
	list <char*>::iterator it;

	//okreslamy sume dlugosci ciagow nazw
	int len=0;
	for (it=l.begin(); it!=l.end(); it++)
	{
		len+=strlen(*it)+1; //+na znak \n
	}

	nameliststr=new char[len+1];
	if (!nameliststr) return "";
	nameliststr[0]=0;

	//dolaczamy kolejne nazwy z funclist do nameliststr
	for (it=l.begin(); it!=l.end(); it++)
	{
#pragma warning(suppress : 4996)
		strcat(nameliststr,*it);
#pragma warning(suppress : 4996)
		strcat(nameliststr,"\n");
	}

	return nameliststr;
}
