//flib.h

#ifndef _FLIB_H_
#define _FLIB_H_

#include <cstring>
#include <list>

using namespace std;


//!Ilosc uzywanych znakow do opisu nazw funkcji i zmiennych
/*!
 *Wchodza w to znaki: A..Z, 0..9 oraz _ (razem 26+10+1=37)
 */
#define FLIB_NCHARS 37

#define FLIB_MAXNAMELEN 64  //!<Maksymalna dlugosc nazwy zmiennej/funkcji


//!Pojedynczy element struktury TNode
typedef struct _TNodeElement
{
	int ref;			//!<Przechowuje adres do funkcji lub zmiennej, albo tez zwykla liczbe
	int flags;			//!<Okresla ilosc parametrow funkcji/sposob interpretowania zmiennej ref
	
	//!Flagi
	enum Flags
	{
		Null,			//!<Wartosc pusta
		Var,			//!<Adres do zmiennej
		Func1,			//!<Adres do funkcji 1-arg
		Func2			//!<Adres do funkcji 2-arg
	};

} TNodeElement;


//!Struktura slownikowa TNode
typedef struct _TNode
{
	TNodeElement *e;				//!<Element wezla
	struct _TNode *n[FLIB_NCHARS];	//!<Tablica wezlow potomnych
} TNode;



//!Klasa reprezentuje strukture slownikowa
/****************//**
*	Zawiera metody dostepowe do zmiennych i funkcji w bibliotece\n
*	czyli dodawanie, usuwanie oraz wyszukiwanie nazw
*****************/
class CFLib
{
private:
	TNode *root;						//korzen struktury

//metody prywatne
	int char2index(int) const;			//zamienia znak na identyfikator w TNode
	int index2char(int) const;			//zamienia identyfikator w TNode na znak

	int str2indextab(const char*,int*,int) const;	//zamienia ciag znakow na ciag identyfikujacy w TNode
										//jesli ciag byl niepoprawny, zwraca 0
										//w przeciwnym wypadku zwraca dlugosc ciagu

	void delnode(TNode *);				//usuwa wezel (funkcja rekurencyjna)


	list <char *> retlist;				//lista, w ktorej przechowywane beda elementy dla ponizszej funkcji
	char *nameliststr;						//lista nazw odzielona znakami nowego wiersza
	list<char *> &funclist(TNode *,				//zwraca liste funkcji o znakach zdefiniowanych w parametrze 2
						  list<char *>&,int=0);	//(funkcja rekurencyjna)
	void eraseretlist();				//czysci retlist oraz zaalokowane ciagi znakow

	//nie dopuszczamy w kodzie
	CFLib(CFLib &);							//konstruktor kopiujacy
	CFLib &operator=(CFLib &);				//podobnie j.w

public:
	CFLib();								//!<Konstruktor, zeruje root i nameliststr

	CFLib(int a) {}						//Konstruktor dodany sztucznie: patrz feval.cpp, linia 43

	~CFLib();							//!<Destruktor, niszczy cale drzewo

	TNodeElement *Add(const char*,int,int);	//!<Dodaje element do biblioteki, zwraca adres do tego elementu

	void Del(const char*);					//!<Usuwa element z biblioteki

	TNodeElement *Find(const char*);		//!<Wyszukuje element w bibliotece o zadanej nazwie

	char *NameList();						//!<Tworzy liste nazw w bibliotece oddzielonych znakami \\n
};

#endif //_FLIB_H_